#include	"MC9S12NE64.h"
#include	"PortDef.h"
#include	"Struct.h"
#include	"Prototype.h"

//
//	Constant
//
#define	BAUD_128000	12
#define BAUD_115200	14
#define BAUD_57600	27
#define	BAUD_38400	41
#define	BAUD_28800	54
#define	BAUD_19200	81
#define	BAUD_14400	109
#define	BAUD_9600		163
#define	BAUD_7200		217

//
//	Prototypes
//
UINT8 GetByte(void);		// Get char from UART
void	PutByte(UINT8);		// Put char for UART
void	InitDevice(void);	// Device init
void	SerialMain(UINT8 *arg);	// Main loop
void	SetupData(void);		// Setup UDP packet
void	DataIndication(void);	// Handle incoming UDP packet
interrupt void Sci0Handler(void);		// SCI0 handler
interrupt void Sci1Handler(void);		// SCI1 handler

//
//	Buffer size constants 
//
#define	ETHERBUF	255		// UDP payload size
#define	RX_SIZE		256		// SCI receive buffer
#define	TX_SIZE		256		// SCI transmit buffer

//
//	UDP Packet data struct
//
typedef	struct 
{
	UINT8	Size;		// Payload size
	UINT8	AckSeq;		// Acknowledged seq number
	UINT8	SeqNum;		// Packet seq number
	UINT8	Pause;		// Pause flag
	UINT8	Time;		// Pause flag

	UINT8	data[ETHERBUF];	// Data buffer
}	PacketBuffer_t;	// Typedef name of UDP packet buffer

//
//	Vars
//
JumpTable HookTable[37] @0x3F91;	// Vector jump table

PacketBuffer_t	*TxPacketBuffer;	// Transmit ether packet buffer
PacketBuffer_t	*RxPacketBuffer;	// Recieve ether packet buffer
UINT8	*RxBuf;			// Receive ring buffer (UART)
UINT8	*TxBuf;			// Transmit ring buffer (UART)

UINT8	RxRingIn;		// Receive ring in pointer (UART)
UINT8	RxRingOut;	// Receive ring out pointer (UART)
UINT8	TxRingIn;		// Transmit ring in pointer (UART)
UINT8	TxRingOut;	// Transmit ring out pointer (UART)
UINT8	StopFlag;		// Stop flag
UINT8 ch;					// Recieved char
UINT8	TxSeqNum;		// My packet seq number
UINT8	RxSeqNum;		// Received packet seq number
UINT8	RxLastNum;	// Last processed packet seq number

UINT16	SrcPort;	// UDP source port number
UINT16	DstPort;	// UDP destination port number
UINT8	Baud;				// Baudrate
UINT8	SciChannel;	// Sci channel(RS-232C or JPEG camera)
UINT8	LoopTime;		// Main loop time
UINT8	PauseCode;	// Pause code
UINT8	FlowCode;		// Flow code
UINT8	AbortCode;	// Abort code

#define	CHANNEL	(SciChannel & 3)
#define	MASTER	(SciChannel & 4)

static const UINT8 BaudRate[8] =	// Baud Rate table
{	BAUD_115200, BAUD_57600, BAUD_38400, BAUD_28800,
	BAUD_19200, BAUD_14400,	BAUD_9600, BAUD_7200};

//
//	SCI0 interrupt handler
//
interrupt void Sci0Handler(void)
{
UINT8 st;		// Work for status & data
	st = SCI0SR1;	// Get status register
	if(st & SCI0SR1_RDRF_MASK)	// Receive data?
	{		// Yes, Serial data has arrived
		ch = SCI0DRL;
		if(RxRingOut != (UINT8)(RxRingIn + 1))	// Ring buffer has space to put?
		{
			RxBuf[RxRingIn++] =	ch;	// Store arrived data to ring buffer
			if((UINT8)(RxRingIn - RxRingOut) > 192)	// Half of ring buffer occupied?
			{
				RTSCOM_OFF;	// Set RTS off for flow control (disable)
				LED1_OFF;		// Show RTS off
			}
		}
		return;	// Don't go to transmit handling(another intterrupt happen when needed)
	}
	if(st & SCI0SR1_TDRE_MASK)	// Trnsmitter is ready to send?
	{	// Yes,then send one char and stop when all data is sent
		if(TxRingIn == TxRingOut)	// Transmit ring buffer is empty?
		{	// Yes,no more data to transmit then disable Tx interrupt
			SCI0CR2_SCTIE = 0;// Clear Tx interrupt bit
			return;	// No more data to send then stop Tx intterrupt and return
		}
		SCI0DRL = TxBuf[TxRingOut++];	// Send one char from transmit ring buffer
	}
}

//
//	SCI1 interrupt handler
//
interrupt void Sci1Handler(void)
{
UINT8 st, ch;		// Work for status & data
	st = SCI1SR1;	// Get status register
	if(st & SCI1SR1_RDRF_MASK)	// Receive data?
	{		// Yes, Serial data has arrived
		ch = SCI1DRL;
		if(RxRingOut != (UINT8)(RxRingIn + 1))	// Ring buffer has space to put?
			RxBuf[RxRingIn++] =	ch;	// Store arrived data to ring buffer
		return;	// Don't go to transmit handling(another intterrupt happen when needed)
	}
	if(st & SCI1SR1_TDRE_MASK)	// Trnsmitter is ready to send?
	{	// Yes,then send one char and stop when all data is sent
		if(TxRingIn == TxRingOut)	// Transmit ring buffer is empty?
		{	// Yes,no more data to transmit then disable Tx interrupt
			SCI1CR2_SCTIE = 0;// Clear Tx interrupt bit
			return;	// No more data to send then stop Tx intterrupt and return
		}
		SCI1DRL = TxBuf[TxRingOut++];	// Send one char from transmit ring buffer
	}
}

//
//	Init timer & UART
//
void InitDevice(void)
{
//	Allocate UART ring buffer for Tx & Rx
	RxBuf = MemoryAlloc(RX_SIZE + TX_SIZE + sizeof(PacketBuffer_t));// Allocate all buffers
	TxBuf = RxBuf + RX_SIZE;					// Allocate Tx ring buffer
	TxPacketBuffer = (PacketBuffer_t*)(TxBuf + TX_SIZE);	// Alloc UDP send buffer
	MemClear(TxPacketBuffer, sizeof(PacketBuffer_t));	// Clear UDP send Packet
	MemClear(&RxRingIn, 9);	// Clear ring pointer & StopFlag & ch & etc same time
//	Init SCI 
	if(CHANNEL)	// SCI channel check
	{											// SCI1 (JPEG camera)
		HookTable[16].adrs = Sci1Handler;	// Set SCI1 interrupt vector
		SCI1BD = BaudRate[Baud];	// Set baud rate of SCI1 (JPEG camera)
	}
	else
	{											// SCI0 (RS-232C)
		HookTable[15].adrs = Sci0Handler;	// Set SCI0 interrupt vector
		SCI0BD = BaudRate[Baud];	// Set baud rate register
		CTSCOM_PE;	// CTS port pull up enable
		RTSCOM_ON;	// Set RTS on
		LED1_ON;	// Show RTS on
	}
}

//
//	Flow control
//
void FlowControl(UINT8 stop)
{
	if(stop)	// Stop requested?
	{
		if(StopFlag == 0)	// Now StopFlag is cleared?
			StopFlag = 1;		// Set flag 1 and will count up in main loop
		TxPacketBuffer->Pause = 1;	// Set pause flag
	}
	else
	{
		StopFlag = 0;			// Clear StopFlag
		TxPacketBuffer->Pause = 0;	// Clear pause flag
	}
}

//
//	Setup UDP packet
//
void SetupData(void)
{
UINT8	index;
	TxPacketBuffer->AckSeq = RxSeqNum;	// Update received seq num for acknowledge
	if(StopFlag && ((PUSHBUTTON3 == PRESSED)	// Check RESTART button (SW3)
	|| (FlowCode && FlowCode == ch)))	// Check flow start code if defined
	{
		FlowControl(0);	// Restart packet flow 		
		return;
	}
	else if(!StopFlag && ((PUSHBUTTON4 == PRESSED)	// Check STOP button (SW4)
	|| (PauseCode && PauseCode == ch)))	// Check pause code if defined
	{
		FlowControl(1);	// Stop packet flow
		return;
	}
	if(TxPacketBuffer->Size)		// Last UDP data confirmed?
		return;					// No then re-send same UDP packet (error retry)
	index = 0;		// Clear store index
	for(;;)	// Data setup loop
	{
		if(RxRingIn == RxRingOut)
			break;				// No more data then exit setup loop
		TxPacketBuffer->data[index]	= RxBuf[RxRingOut++];	// Store char to packet buffer
		if(CHANNEL == 0	// SCI0 mode?
			&& (UINT8)(RxRingIn - RxRingOut) < 64)// Data amout is less than 1/4?
		{
			RTSCOM_ON;	// set RTS on for flow control (flow enable)
			LED1_ON;			// Show RTS is ON
		}
		TxPacketBuffer->Size = ++index;	// Add length
		if(index == ETHERBUF)	// Pay load size limit?
			break;	// Exit data setup loop when payload is full
	}
	if(index)	// Data exists?
		LED3_ON;				// Light send data LED
	if(++TxSeqNum == 0)	// Increment Packet seq number for new UDP packet
		TxSeqNum++;			// If it's 0 then force to 1
	TxPacketBuffer->SeqNum = TxSeqNum;	// Set Send Seq num
}

//
//	Handle incoming UDP packet
//
void DataIndication(void)
{
UINT8	i;	// Loop counter work
	if(RxPacketBuffer->AckSeq == TxSeqNum)	// Ack seq num matches sent seq num?
		TxPacketBuffer->Size = 0;	// Clear size for next packet send(acknowledged)
	if(StopFlag == 0 || StopFlag == 20)	// Test not paused or paused
		FlowControl(RxPacketBuffer->Pause);	// Packet flow control
	else																	// case of StopFlag is 1 from 19
		TxPacketBuffer->Pause = RxPacketBuffer->Pause;	// Reply pause code to sender
	if(!CHANNEL && CTSCOM	// Check CTS off for flow control
	|| (UINT8)(TxRingIn - TxRingOut) + RxPacketBuffer->Size >= TX_SIZE)
// Check recieved payload will fit Tx ring buffer?(too big payload?)
	{	// Noway to put received data to UART transmit ring buffer
		RxSeqNum = 0; // Force request resend UDP buffer(CTS off or Tx ring is full)
		return;		// Return (not set data to Tx ring buffer)
	}
	RxSeqNum = RxPacketBuffer->SeqNum; // Save recieved packet seq number
	if(RxSeqNum == RxLastNum)	// Already processed last time?
		return;			// Return if it was duplicate packet (dump packet)
	RxLastNum = RxSeqNum;	// Save processed packet seq num
	if(RxPacketBuffer->Size)	// Received packet has payload?
	{	// Yes then put UDP payload to Tx ring buffer for UART transmittion
		for(i = 0; i < RxPacketBuffer->Size; i++)	// Send payload data to Tx ring
			TxBuf[TxRingIn++] = RxPacketBuffer->data[i];	// Put received char to Tx ring buffer
		LED2_ON;	// Light receive data LED 
		if(CHANNEL)	// Switch by SCI channel
			SCI1CR2_SCTIE = 1;	// Start SCI1 transmit
		else
			SCI0CR2_SCTIE = 1;	// Start SCI0 transmit
	}
}

//
//	Network serial cable main
//	Arg structure as follows
//	IP(4byte),SrcPort(2byte),DstPort(2byte),Baud(1byte),
//	Flow && Channel(1byte),Interval(1byte),
//	PauseCode(1byte),FlowCode(1byte),AbortCode(1byte)
//
void SerialMain(UINT8 *arg)
{
UINT8	i, Soc, Timer;
//	Setup parameters
	BufCopy(&SrcPort, arg + 4, 10);// Set SrcPort,DstPort,Baud,SciChannel,LoopTime,PauseCode,FlowCode
	InitDevice();		// Init timer & UART and set baud rate
	Soc = CreateSocket(0);		// Create UDP socket
	(void)Bind(Soc, SrcPort, 0);		// Set source port number
	Timer = CreateTimer(1, 0, 1, 0, 0);	// Create timer
// Start main loop
	if(MASTER)	// Master mode?
		LoopTime--;	// Then reduce loop time to make +/- margine
	for(;;)	// Repeat forever
	{	// Transmit UDP and wait UDP reply 
		LED2_OFF;					// Turn off recieved LED
		LED3_OFF;					// Turn off send data LED
		LED4_OFF;					// Turn off link up LED
		if(AbortCode && AbortCode == ch)	// Check abort
			break;		// Exit main loop
		(void)CreateTimer(1, 0, LoopTime, Timer, 0);	// Create timer
		SetupData();		// Setup UDP packet
		if(StopFlag < 20)	// Send packet when StopFlag == 0 or les then 20
		{
			(void)SendTo(Soc, *(UINT32*)arg, DstPort,	// Send UDP packet 
				(UINT8*)TxPacketBuffer, TxPacketBuffer->Size + 5);// Pointer & length
			if(StopFlag)		// If StopFlag is set
				StopFlag++;		// Count up Stop flag (1 to 20 count up)
		}
		i = 0;	// Clear timer adjust value
		while(GetTimerCount(Timer))	// Wait reply packet while LoopTime
		{
			if(RecvFrom(Soc, 0) >= 0)	// Check UDP packet arrival
			{	// Yes,UDP data has arrived
				RxPacketBuffer = (PacketBuffer_t*)GetReceiveBuffer(Soc, 1);// Get UDP data buffer
				if((UINT16)GetSenderIP(Soc) == *(UINT16*)(arg + 2)	// Check IP adrs
					&& GetSenderPort(Soc) == SrcPort)	        // Check port number
				{
					LED4_ON;	// Light link up LED
					TxPacketBuffer->Time = (UINT8)GetTimerCount(Timer);	// Reply received time for SYNC
					i = RxPacketBuffer->Time;	// Get arrival time of my packet
					DataIndication();	// Handle incoming UDP packet
					MemoryFree(RxPacketBuffer);	// Release recieved UDP data buffer
				}
				else
					MemoryFree(RxPacketBuffer);	// Dump recieved UDP data buffer
			}
			SystemSleep();	// Allow another thread to execute their task
		}
		if(MASTER && i--)	// Only slave receievd packe before timeout then adjust timing
		{
			while(i)	// If arrval time has value (means early arrival)
			{
				Sleep(1);	// Insert delay to cancel early arrival
				i--;		// Decrement timer value 
				SystemSleep();	// Allow another thread to execute their task
			}
		}	// End of timing adjustment (Only for Master)
	}		// End of transfer main loop
	(void)CreateTimer(0, 0, 0, Timer, 0);	// Create timer
	CloseSocket(Soc);		// Close socket
	MemoryFree(RxBuf);	// Release allocated buffer
}
